import os
import json
import zlib
import time
import platform
import requests
from urllib.parse import urljoin
import sys
import re
import mimetypes

root_path = os.path.join(os.path.dirname(__file__), "..", "..", "examples/Binary/mac")
json_file = os.path.join(root_path, "file_structure.json")
base_url = "https://builds.luckyrobots.xyz/"

def get_os_type():
    os_type = platform.system().lower()
    if os_type == "darwin":
        return "mac"
    elif os_type == "windows":
        return "win"
    elif os_type == "linux":
        return "linux"
    else:
        raise ValueError(f"Unsupported operating system: {os_type}")

def calculate_crc32(file_path):
    with open(file_path, 'rb') as file:
        crc32 = 0
        while True:
            data = file.read(65536)  # Read in 64kb chunks
            if not data:
                break
            crc32 = zlib.crc32(data, crc32)
    return crc32 & 0xFFFFFFFF  # Ensure unsigned 32-bit integer

def scan_directory(root_path):
    file_structure = []
    
    for dirpath, dirnames, filenames in os.walk(root_path):
        # Add directories
        for dirname in dirnames:
            dir_path = os.path.join(dirpath, dirname)
            relative_path = os.path.relpath(dir_path, root_path)
            file_structure.append({
                "path": relative_path,
                "type": "directory",
                "size": 0,  # Directories don't have a size in this context
                "mtime": os.path.getmtime(dir_path),
                "crc32": 0,
                "mime_type": "directory"
            })
        
        # Add files
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            relative_path = os.path.relpath(file_path, root_path)
            file_stat = os.stat(file_path)
            
            # Guess the file type using mimetypes
            file_type, encoding = mimetypes.guess_type(file_path)
            if file_type is None:
                file_type = "application/octet-stream"  # Default to binary if type can't be guessed
            
            file_structure.append({
                "path": relative_path,
                "crc32": calculate_crc32(file_path),
                "size": file_stat.st_size,
                "mtime": file_stat.st_mtime,
                "type": "file",
                "mime_type": file_type
            })
    
    return file_structure

def save_json(data, filename):
    with open(filename, 'w') as f:
        json.dump(data, f, indent=2)

def load_json(filename):
    with open(filename, 'r') as f:
        return json.load(f)

def clean_path(path):
    # Remove ['children'] and quotes, replace '][' with '/'
    cleaned = re.sub(r"\['children'\]", "", path).replace("']['", "/").strip("[]'")
    # Replace remaining single quotes with nothing
    cleaned = cleaned.replace("'", "")
    # Split the path and the attribute that changed
    parts = cleaned.rsplit('/', 1)
    return parts[0], parts[1] if len(parts) > 1 else None

def compare_structures(json1, json2):
    dict1 = {item['path']: item for item in json1}
    dict2 = {item['path']: item for item in json2}

    result = []

    # Check for new and modified items
    for path, item in dict2.items():
        if path not in dict1:
            item['change_type'] = 'new_file'
            result.append(item)
        elif item['type'] == 'file' and item['crc32'] != dict1[path]['crc32']:
            item['change_type'] = 'modified'
            result.append(item)
        # Unchanged items are not added to the result

    # Check for deleted items
    for path, item in dict1.items():
        if path not in dict2:
            item['change_type'] = 'deleted'
            result.append(item)

    # Remove the item with "path": "hashmap.json" from the result
    result = [item for item in result if item['path'] != "hashmap.json"]
    # save_json(result, "changes.json")
    return result


def scan_server(server_path):
    
    mac_path = os.path.join(server_path, "mac")
    win_path = os.path.join(server_path, "win")
    linux_path = os.path.join(server_path, "linux")
    
    mac_structure = scan_directory(mac_path)
    win_structure = scan_directory(win_path)
    linux_structure = scan_directory(linux_path)
    
    save_json(mac_structure, os.path.join(server_path, "mac/hashmap.json"))
    save_json(win_structure, os.path.join(server_path, "win/hashmap.json"))
    save_json(linux_structure, os.path.join(server_path, "linux/hashmap.json"))

def check_updates(root_path):
    # Determine the operating system
    os_type = get_os_type()
    global base_url
    # Set the base URL

    # Construct the URL based on the operating system
    os_type = get_os_type()
    url = urljoin(base_url, f"{os_type}/hashmap.json")

    # Download the JSON file
    try:
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for HTTP errors
        server_structure = response.json()
    except requests.RequestException as e:
        print(f"Error downloading JSON file: {e}")
        server_structure = None

    if server_structure is None:
        print("Using local scan as no remote file could be downloaded.")
        server_structure = {}

    # Scan the directory and create a new file structure
    client_structure = scan_directory(root_path)

    # If a previous scan exists, compare and show changes
    if server_structure:
        changes = compare_structures(client_structure, server_structure)
        
        # for future debugging
        # # Write the flat JSON to a file
        # with open('changes.json', 'w') as f:
        #     json.dump(changes, f, indent=2)
        
        # print(f"Changes have been written to changes.json")
    else:
        print("No server structure available for comparison.")

    # for future debugging
    # # Save the new structure
    # save_json(client_structure, "./client_structure.json")
    
    return changes

if __name__ == "__main__":
    lr_server_root = None
    
    for arg in sys.argv:
        if arg.startswith('--lr-server-root='):
            lr_server_root = arg.split('=', 1)[1]
            break

    if lr_server_root:
        # this is being used as a cron job on the main server to keep the client file structures up to date
        print(f"Scanning server at {lr_server_root}")
        scan_server(lr_server_root)



from .event_emitter import EventEmitter

event_emitter = EventEmitter()

def on(event):
    def decorator(callback):
        event_emitter.on(event, callback)
        return callback
    return decorator


import os
import json
import time
from watchdog.events import FileSystemEventHandler


class Handler(FileSystemEventHandler):
    file_num = 0
    image_stack = {}
    send_bytes = False
    emit_counter = 0
    lucky_robots = None  # Initialize this as None
    

    @classmethod
    def set_send_bytes(cls, value):
        cls.send_bytes = value

    @classmethod
    def set_lucky_robots(cls, lr):
        cls.lucky_robots = lr

    @staticmethod
    def on_created(event):
        if not event.is_directory:
            Handler.process_file(event.src_path, event.event_type)

    @staticmethod
    def get_file_name(file_path):
        return os.path.basename(file_path)

    @staticmethod
    def on_modified(event):
        if not event.is_directory:
            Handler.process_file(event.src_path, event.event_type)

    @staticmethod
    def process_file(file_path, event_type):
        file = Handler.get_file_name(file_path)
        current_file_num = int(file.split('_')[0]) if file.split('_')[0].isdigit() else Handler.file_num
        
        if current_file_num == Handler.file_num:
            Handler.add_file(file_path)
        else:
            # emit.counter 5 to give file watcher some warmup time.
            if len(Handler.image_stack) > 0 and Handler.emit_counter > 5:
                if Handler.lucky_robots is not None:
                    Handler.lucky_robots.message_received_sync("robot_output", Handler.image_stack)
                else:
                    print("Warning: lucky_robots is not set in Handler class")
            Handler.emit_counter += 1
            Handler.file_num = current_file_num
            Handler.add_file(file_path)
            
            # print(Handler.image_stack)
        # print(f"Processed file from {event_type} event: {file_path}")

    @staticmethod
    def add_file(file_path):
        file = Handler.get_file_name(file_path)
        file_bytes = None
        file_type = file.split('_', 1)[1].rsplit('.', 1)[0]
        if file_path not in Handler.image_stack:
            if file.endswith('.txt'):
                try:
                    with open(file_path, 'r') as f:
                        file_content = f.read()
                        file_bytes = json.loads(file_content)
                except json.JSONDecodeError:
                    print(f"Error decoding JSON from {file_path}")
                    file_bytes = {}
                except IOError as e:
                    print(f"Error reading file {file_path}: {e}")
                    file_bytes = {}
            else:
                if Handler.send_bytes:
                    file_bytes = Handler._read_file_with_retry(file_path)
                else:
                    file_bytes = None
            
            Handler.image_stack[file_type] = {"file_path": file_path, "contents": file_bytes}        


    @staticmethod
    def on_deleted(event):
        if event.is_directory:
            return None
        else:
            # print(f"Received deleted event - {event.src_path}")
            pass

    @staticmethod
    def _read_file_with_retry(file_path, retries=5, delay=1):
        for attempt in range(retries):
            try:
                with open(file_path, 'rb') as f:
                    content = f.read()
                    return content
            except (PermissionError, FileNotFoundError) as e:
                print(f"Attempt {attempt + 1}: Failed to read {file_path} - {e}")
                time.sleep(delay)
        else:
            print(f"Failed to read {file_path} after {retries} attempts")



import os
import tempfile
import psutil
import subprocess
import platform
import atexit
import sys

LOCK_FILE = os.path.join(tempfile.gettempdir(), 'luckeworld_lock')

def is_luckyworld_running():
    # Check for the lock file
    if os.path.exists(LOCK_FILE):
        with open(LOCK_FILE, 'r') as f:
            pid = int(f.read().strip())
        if psutil.pid_exists(pid):
            # Double-check if the process is actually LuckEWorld
            try:
                process = psutil.Process(pid)
                if "LuckEWorld" in process.name() or "luckyrobots" in process.name():
                    return True
            except psutil.NoSuchProcess:
                pass  # Process doesn't exist, continue to remove lock file
        
        # If we reach here, the lock file is stale
        remove_lock_file()
    
    # Check for any running LuckEWorld processes
    for proc in psutil.process_iter(['name']):
        if "LuckEWorld" in proc.info['name'] or "luckyrobots" in proc.info['name']:
            create_lock_file(proc.pid)
            return True
    
    return False

def create_lock_file(pid):
    with open(LOCK_FILE, 'w') as f:
        f.write(str(pid))

def remove_lock_file():
    if os.path.exists(LOCK_FILE):
        os.remove(LOCK_FILE)

def run_luckyworld_executable(directory_to_watch):
    # Determine the correct path based on the operating system
    if platform.system() == "Darwin":  # macOS
        executable_path = os.path.join(directory_to_watch, "..","..","..","MacOS", "luckyrobots")
    elif platform.system() == "Linux":  # Linux
        executable_path = os.path.join(directory_to_watch, "..","..","luckyrobots.sh")
    else:  # Windows or other platforms
        executable_path = os.path.join(directory_to_watch, "..","..","luckyrobots.exe")        

    # Check if the executable exists
    if not os.path.exists(executable_path):
        print(f"Error: Executable not found at {executable_path}")
        sys.exit(1)
        return

    try:
        # Set execute permissions
        os.chmod(executable_path, 0o755)
        print("running executable at:", executable_path)
        # Check if --lr-verbose flag is used
        verbose = "--lr-verbose" in sys.argv

        # Run the executable as a detached process
        if platform.system() == "Windows":
            # For Windows
            DETACHED_PROCESS = 0x00000008
            process = subprocess.Popen(
                executable_path,
                creationflags=DETACHED_PROCESS,
                close_fds=True,
                stdout=subprocess.DEVNULL if not verbose else None,
                stderr=subprocess.DEVNULL if not verbose else None
            )
        else:
            # For Unix-based systems (macOS, Linux)
            process = subprocess.Popen(
                executable_path,
                start_new_session=True,
                stdout=subprocess.DEVNULL if not verbose else None,
                stderr=subprocess.DEVNULL if not verbose else None
            )
        
        # Create lock file with the new process ID
        create_lock_file(process.pid)
        
        if verbose:
            print("LuckEWorld application started successfully as an independent process.")
    except subprocess.CalledProcessError as e:
        print(f"Error: Failed to start LuckEWorld application. {e}")
    except PermissionError as e:
        print(f"Error: Permission denied. Unable to set execute permissions. {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")

# Ensure lock file is removed if the script exits unexpectedly
atexit.register(remove_lock_file)


from watchdog.observers import Observer
from .handler import Handler

class Watcher:
    def __init__(self, directory_path):
        self.observer = Observer()
        self.directory_path = directory_path
    def run(self):
        event_handler = Handler()
        self.observer.schedule(event_handler, self.directory_path, recursive=False)
        self.observer.start()
        try:
            while self.observer.is_alive():
                self.observer.join(1)
        except KeyboardInterrupt:
            self.observer.stop()
        self.observer.join()


import asyncio
import os
import time
import sys
import json
import threading
import tempfile
import platform
from fastapi import FastAPI, WebSocket
import uvicorn
import logging
import random
import signal

from .watcher import Watcher
from .handler import Handler
from .download import check_binary
from .library_dev import library_dev
from .run_executable import is_luckyworld_running, run_luckyworld_executable



LOCK_FILE = os.path.join(tempfile.gettempdir(), 'luckeworld_lock')


app = FastAPI()


class LuckyRobots:
    delegate_object = None
    websocket = None
    port = 3000
    host = "0.0.0.0"
    
    
    @staticmethod
    def set_delegate_object(delegate_object):
        LuckyRobots.set_delegate_object(delegate_object)

    @staticmethod
    def get_websocket():
        return LuckyRobots.websocket

    @staticmethod
    async def message_received(message,data=None):
        # print("LuckyRobots message_received", message)
        if LuckyRobots.receiver_function is not None:
            await LuckyRobots.receiver_function(message,data)

    @staticmethod
    def message_received_sync(message,data=None):
        # print("LuckyRobots message_received_sync", message)
        asyncio.run(LuckyRobots.message_received(message,data))
        
    @staticmethod
    def message_receiver(func):        
        # Check if the function is asynchronous
        if not asyncio.iscoroutinefunction(func):
            # Get the function's name and try to get its location
            func_name = func.__name__
            try:
                func_location = f" in {func.__code__.co_filename}"
            except AttributeError:
                func_location = ""
            
            error_message = (
                f"Error: The message receiver function '{func_name}'{func_location} "
                "must be asynchronous. Use 'async def' to define the function."
            )
            print(error_message)
            LuckyRobots.run_exit_handler()
        
        """Decorator to set the receiver function"""
        LuckyRobots.receiver_function = func
        return func


    @staticmethod
    async def send_message(message):
        if LuckyRobots.websocket is not None:
            await LuckyRobots.websocket.send_text(message)
        else:
            print("WebSocket connection is not established yet.", LuckyRobots.websocket)


    @staticmethod
    async def send_commands(commands):
        def get_random_int():
            return random.randint(0, 1000000)
        
        if isinstance(commands, list):
            instructions = {"LuckyCode": []}
        
            for command in commands:
                instructions["LuckyCode"].append({
                    "ID": str(command["id"]) if isinstance(command, dict) and 'id' in command else str(get_random_int()),
                    "code": str(command["code"]) if isinstance(command, dict) and 'code' in command else str(command),
                    "time": str(int(time.time() * 1000)),
                    "callback": "off"
                })
        else:
            instructions = commands 

        print("instructions", instructions)
        
        # Add this line to actually send the commands
        await LuckyRobots.send_message(json.dumps(instructions))
        
    @staticmethod
    @app.websocket("/ws")
    async def websocket_endpoint(websocket: WebSocket):

        try:
            await websocket.accept()
            print("WebSocket connection established")
            LuckyRobots.websocket = websocket
            # asyncio.run(event_emitter.emit("start",ws))
            
            
        except Exception as e:
            print(f"Failed to establish WebSocket connection: {e}")
            return


        try:
            while True:
                data = await websocket.receive_text()
                await LuckyRobots.message_received(data)
                # handle_event(data)
        except Exception as e:
            print(f"WebSocket error: {e}")
        finally:
            ws = None
            print("WebSocket connection closed")

    def run_server():
        # event_emitter.emit("message", "running server on port 3000")
        # asyncio.get_event_loop().run_until_complete(event_emitter.emit("message", "running server on port 3000"))
        # Configure logging
        logging.getLogger("uvicorn.access").setLevel(logging.WARNING)
        
        # Run the server with custom log config
        uvicorn.run(app, host=LuckyRobots.host, port=LuckyRobots.port, log_level="warning")

    @staticmethod
    def start(binary_path=None, send_bytes=False):
        

        if binary_path is None:
            binary_path = check_binary("./Binary")    
        if not os.path.exists(binary_path):
            print(f"I couldn't find the binary at {binary_path}, are you sure it's running and capture mode is on?")
            os._exit(1)
        



        if sys.platform == "darwin":  # macOS
            directory_to_watch = os.path.join(binary_path, "luckyrobots.app", "Contents", "UE", "luckyrobots", "robotdata")
        else:  # Windows and other platforms
            directory_to_watch = os.path.join(binary_path, "luckyrobots", "robotdata")
        
        # Create the directory if it doesn't exist
        os.makedirs(directory_to_watch, exist_ok=True)
            
        Handler.set_send_bytes(send_bytes)
        Handler.set_lucky_robots(LuckyRobots)  # Set the LuckyRobots class in Handler
        
        if is_luckyworld_running():
            print("LuckyRobots is already running. Skipping launch.")
        else:
            # Start the LuckEWorld executable
            if "--lr-no-executable" not in sys.argv:
                run_luckyworld_executable(directory_to_watch)

        library_dev() 

        # Run the server in a separate thread
        server_thread = threading.Thread(target=LuckyRobots.run_server, daemon=True)
        server_thread.start()

        print("*" * 60)
        print("                                                                                ")
        print("                                                                                ")
        print("▄▄▌  ▄• ▄▌ ▄▄· ▄ •▄  ▄· ▄▌▄▄▄        ▄▄▄▄·       ▄▄▄▄▄.▄▄ · ")
        print("██•  █▪██▌▐█ ▌▪█▌▄▌▪▐█▪██▌▀▄ █·▪     ▐█ ▀█▪▪     •██  ▐█ ▀. ")
        print("██▪  █▌▐█▌██ ▄▄▐▀▀▄·▐█▌▐█▪▐▀▀▄  ▄█▀▄ ▐█▀▀█▄ ▄█▀▄  ▐█.▪▄▀▀▀█▄")
        print("▐█▌▐▌▐█▄█▌▐███▌▐█.█▌ ▐█▀·.▐█•█▌▐█▌.▐▌██▄▪▐█▐█▌.▐▌ ▐█▌·▐█▄▪▐█")
        print(".▀▀▀  ▀▀▀ ·▀▀▀ ·▀  ▀  ▀ • .▀  ▀ ▀█▄▀▪·▀▀▀▀  ▀█▄▀▪ ▀▀▀  ▀▀▀▀ ")
        print("                                                                                ")
        print("                                                                                ")
        if platform.system() == "Darwin":
            print("*" * 60)
            print("For macOS users:")
            print("Please be patient. The application may take up to a minute to open on its first launch.")
            print("If the application doesn't appear, please follow these steps:")
            print("1. Open System Settings")
            print("2. Navigate to Privacy & Security")
            print("3. Scroll down and click 'Allow' next to the 'luckyrobots' app")
            print("*" * 60)    
        print("Lucky Robots application started successfully.")
        print("To move the robot from your python code, choose a level on the game, and tick the HTTP checkbox.")
        print("To receive the camera feed from your python code, choose a level on the game, and tick the Capture checkbox.")    
        print("*" * 60)
        
        # Check if the system is macOS

        # Set up a signal handler for SIGINT (Ctrl+C)
        

        def sigint_handler(signum, frame):
            print("\nCtrl+C pressed. Running exit handler...")
            LuckyRobots.run_exit_handler(ctrlc_pressed=True)

        signal.signal(signal.SIGINT, sigint_handler)


        watcher = Watcher(directory_to_watch)
        watcher.run()
            
    @staticmethod     
    def run_exit_handler(ctrlc_pressed=False):

        def signal_handler(sig, frame):
            print("\nExiting gracefully...")
            # Perform any cleanup if necessary
            exit(0)

        # Register the signal handler for SIGINT (Ctrl+C)
        signal.signal(signal.SIGINT, signal_handler)

        # Set up a loop to keep the program running
        
        if ctrlc_pressed:
            print("Exiting...")
            sys.exit(0)
        else:
            print("Press Enter to exit the program")
            try:
                while True:
                    # Use input() to allow the program to be interrupted by Enter key
                    user_input = input()
                    if user_input:
                        print("Exiting...")
                        break
            except KeyboardInterrupt:
                # This block will be executed if Ctrl+C is pressed
                pass



# Remove this line if remove_lock_file is not defined or used:
# atexit.register(remove_lock_file)

# Export the necessary functions and classes
__all__ = ['LuckyRobots']



import time

class Test:
    
    
    @staticmethod
    def receiver(func):
        Test.receiver_function = func
        return func

    @staticmethod
    def message_received(message):
        print("Test message_received", message)
        if Test.receiver_function is not None:
            Test.receiver_function(message)
        else:
            print("Test message_received no receiver function") 
            
            
            
            
@Test.receiver
def receiver(message):
    print("receiver ", message)
    pass

while True:
    time.sleep(1)
    Test.message_received("hello from python")


from .core import LuckyRobots
from .check_updates import check_updates            
# Expose static methods
start = LuckyRobots.start
send_message = LuckyRobots.send_message
send_commands = LuckyRobots.send_commands
message_receiver = LuckyRobots.message_receiver

# Export the necessary functions and classes
__all__ = ['LuckyRobots', 'start', 'send_message', 'send_commands', 'message_receiver', 'check_updates']